/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

/*
 * ==================================================================
 * This file contains the source for a simple MPI process manager
 * used by MVICH.
 * It simply collects the arguments and execs either RSH or SSH
 * to execute the processes on the remote (or local) hosts.
 * Some critical information is passed to the remote processes
 * through environment variables using the "env" utility. 
 *
 * The information passed through the environment variables is:
 *  MPIRUN_HOST = host running this mpirun_rsh command
 *  MPIRUN_PORT = port number mpirun_rsh is listening on for TCP connection
 *  MPIRUN_RANK = numerical MPI rank of remote process
 *  MPIRUN_NPROCS = number of processes in application
 *  MPIRUN_ID   = pid of the mpirun_rsh process
 *
 * The remote processes use this to establish TCP connections to
 * this mpirun_rsh process.  The TCP connections are used to exchange
 * address data needed to establish the VI connections.
 * The TCP connections are also used for a simple barrier syncronization
 * at process termination time.
 *
 * MVICH allows for the specification of certain tuning parameters
 * at run-time.  These parameters are read by mpirun_rsh from a
 * file given on the command line.  Currently, these parameters are
 * passed to the remote processes through environment variables, but 
 * they could be sent as a string over the TCP connection.  It was
 * thought that using environment variables might be more portable
 * to other process managers.
 * ==================================================================
 */


#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#ifdef MAC_OSX
#include <sys/wait.h>
#else
#include <wait.h>
#endif
#include <assert.h>
#include <fcntl.h>
#include "pmgr_client.h"

typedef enum {
    P_NOTSTARTED,
    P_STARTED,
    P_CONNECTED,
    P_DISCONNECTED,
    P_RUNNING,
    P_FINISHED,
    P_EXITED
} process_state;

typedef struct {
    char *hostname;
    int pid;
    int control_socket;
    process_state state;
    int port;
} process;

#define RUNNING(i) ((plist[i].state == P_STARTED ||                 \
            plist[i].state == P_CONNECTED ||                        \
            plist[i].state == P_RUNNING) ? 1 : 0)

/* other information: a.out and rank are implicit. */

process *plist;
int nprocs;
int aout_index, port;
#define MAX_WD_LEN 256
char wd[MAX_WD_LEN];            /* working directory of current process */
#define MAX_HOST_LEN 256
char mpirun_host[MAX_HOST_LEN]; /* hostname of current process */
/* xxx need to add checking for string overflow, do this more carefully ... */


#define COMMAND_LEN 2000


void cleanup_handler(int);
void nostop_handler(int);
void alarm_handler(int);
void child_handler(int);
void usage(void);
int start_process(int i, char *command_name, char *env);
void cleanup(void);
char *skip_white(char *s);
char *read_param_file(char *paramfile);
void process_termination(void);
int set_fds(fd_set * rfds, fd_set * efds);

#define RSH_CMD	"/usr/bin/rsh"
#define SSH_CMD	"/usr/bin/ssh"
#ifdef USE_SSH
int use_rsh = 0;
#else
int use_rsh = 1;
#endif

#define SSH_ARG "-q"

#define SH_NAME_LEN	(128)
char sh_cmd[SH_NAME_LEN];

#define XTERM "/usr/X11R6/bin/xterm"
#define ENV_CMD "/usr/bin/env"
#define TOTALVIEW_CMD "/usr/totalview/bin/totalview"

static struct option option_table[] = {
    {"np", required_argument, 0, 0},
    {"debug", no_argument, 0, 0},
    {"xterm", no_argument, 0, 0},
    {"hostfile", required_argument, 0, 0},
    {"paramfile", required_argument, 0, 0},
    {"show", no_argument, 0, 0},
    {"rsh", no_argument, 0, 0},
    {"ssh", no_argument, 0, 0},
    {"help", no_argument, 0, 0},
    {"v", no_argument, 0, 0},
    {"tv", no_argument, 0, 0},
    {0, 0, 0, 0}
};

static void show_version(void)
{
    fprintf(stderr,"OSU MVAPICH VERSION 0.9.5-SingleRail\n");
}


int debug_on = 0, xterm_on = 0, show_on = 0;
int param_debug = 1;
int use_totalview = 0;
char * mpirun_processes;
static char display[200];

static void get_display_str()
{
    char *p;
    char str[200];

    if ( (p = getenv( "DISPLAY" ) ) != NULL ) {
	strcpy(str, p ); /* For X11 programs */  
	sprintf(display,"DISPLAY=%s",str); 	
    }
}

int main(int argc, char *argv[])
{
    int i, s, s1, c, option_index;
    int hostfile_on = 0;
#define HOSTFILE_LEN 256
    char hostfile[HOSTFILE_LEN + 1];
    int paramfile_on = 0;
#define PARAMFILE_LEN 256
    char paramfile[PARAMFILE_LEN + 1];
    FILE *hf;
    struct sockaddr_in sockaddr;
    unsigned int sockaddr_len = sizeof(sockaddr);
    int addrlen, global_addrlen = 0;

    char *env = "\0";
    int tot_nread = 0;

    int *alladdrs = NULL;
    char *alladdrs_char = NULL; /* for byte location */
    int *out_addrs;
    int out_addrs_len;
    int j;

    char *allpids=NULL;
    int pidlen, pidglen;

    char command_name[COMMAND_LEN];
    char command_name_tv[COMMAND_LEN];
    char totalview_cmd[200];
    char *tv_env;

    int    hostname_len = 0;

    totalview_cmd[199] = 0;
    display[0]='\0';	
     
    /* mpirun [-debug] [-xterm] -np N [-hostfile hfile | h1 h2 h3 ... hN] a.out [args] */

    do {
        c = getopt_long_only(argc, argv, "+", option_table, &option_index);
        switch (c) {
        case '?':
        case ':':
            usage();
            break;
        case EOF:
            break;
        case 0:
            switch (option_index) {
            case 0:
                nprocs = atoi(optarg);
                if (nprocs < 1)
                    usage();
                break;
            case 1:
                debug_on = 1;
                xterm_on = 1;
                break;
            case 2:
                xterm_on = 1;
                break;
            case 3:
                hostfile_on = 1;
                strncpy(hostfile, optarg, HOSTFILE_LEN);
                if (strlen(optarg) >= HOSTFILE_LEN - 1)
                    hostfile[HOSTFILE_LEN] = '\0';
                break;
            case 4:
                paramfile_on = 1;
                strncpy(paramfile, optarg, PARAMFILE_LEN);
                if (strlen(optarg) >= PARAMFILE_LEN - 1) {
                    paramfile[PARAMFILE_LEN] = '\0';
                }
                break;
            case 5:
                show_on = 1;
                break;
            case 6:
                use_rsh = 1;
                break;
            case 7:
                use_rsh = 0;
                break;
            case 8:
            case 9:
                show_version();
                exit(0);
                break;
 	    case 10:
 		use_totalview = 1;
		debug_on = 1;
 		tv_env = getenv("TOTALVIEW");
 		if(tv_env != NULL)
 		{
		    strcpy(totalview_cmd,tv_env);	
		    /*fprintf(stderr, "TOTALVIEW: %s\n",totalview_cmd);*/
 		} else {
		    fprintf(stderr,
			    "TOTALVIEW env is NULL, use default: %s\n", 
			    TOTALVIEW_CMD);
		    sprintf(totalview_cmd, "%s", TOTALVIEW_CMD);
 		}	
  		break;
            case 11:
                usage();
                exit(0);
                break;
            default:
                fprintf(stderr, "Unknown option\n");
                usage();
                break;
            }
            break;
        default:
            fprintf(stderr, "Unreachable statement!\n");
            usage();
            break;
        }
    } while (c != EOF);

    if (!hostfile_on) {
        /* get hostnames from argument list */
        if (argc - optind < nprocs + 1) {
            fprintf(stderr, "Without hostfile option, hostnames must be "
                    "specified on command line.\n");
            usage();
            exit(1);
        }
        aout_index = nprocs + optind;
    } else {
        aout_index = optind;
    }

    if (paramfile_on) {
        /* construct a string of environment variable definitions from
         * the entries in the paramfile.  These environment variables
         * will be available to the remote processes, which
         * will use them to over-ride default parameter settings
         */
        env = read_param_file(paramfile);
    }

    plist = malloc(nprocs * sizeof(process));
    if (plist == NULL) {
        perror("malloc");
        exit(1);
    }

    for (i = 0; i < nprocs; i++) {
        plist[i].state = P_NOTSTARTED;
    }

    /* grab hosts from command line or file */

    if (hostfile_on) {
        hf = fopen(hostfile, "r");
        if (hf == NULL) {
            fprintf(stderr, "Can't open hostfile %s\n", hostfile);
            perror("open");
            exit(1);
        }

        for (i = 0; i < nprocs; i++) {
            char line[100];
            if (fgets(line, 100, hf) != NULL) {
                int len = strlen(line);
                if (line[len - 1] == '\n')
                    line[len - 1] = '\0';
                plist[i].hostname = strdup(line);
		hostname_len = hostname_len > len ? hostname_len : len;
            } else {
                fprintf(stderr, "End of file reached on "
                        "hostfile at %d of %d hostnames\n", i, nprocs);
                exit(1);
            }
        }
        fclose(hf);
    } else {
        for (i = 0; i < nprocs; i++) {
            plist[i].hostname = argv[optind + i];
	    hostname_len = hostname_len > strlen(plist[i].hostname) ?
		    hostname_len : strlen(plist[i].hostname); 
        }
    }

    mpirun_processes = (char*)malloc(sizeof(char) * nprocs * (hostname_len + 4));

    if (!mpirun_processes) {
        perror("malloc");
        exit(1);
    } else { 
	memset(mpirun_processes, 0, nprocs * (hostname_len + 4));
    }

    for (i = 0; i < nprocs; ++i) {
	strcat(mpirun_processes, plist[i].hostname);
	strcat(mpirun_processes, ":");
    }

    getcwd(wd, MAX_WD_LEN);
    gethostname(mpirun_host, MAX_HOST_LEN);

    get_display_str();

    s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (s < 0) {
        perror("socket");
        exit(1);
    }
    sockaddr.sin_addr.s_addr = INADDR_ANY;
    sockaddr.sin_port = 0;
    if (bind(s, (struct sockaddr *) &sockaddr, sockaddr_len) < 0) {
        perror("bind");
        exit(1);
    }

    if (getsockname(s, (struct sockaddr *) &sockaddr, &sockaddr_len) < 0) {
        perror("getsockname");
        exit(1);
    }

    port = (int) ntohs(sockaddr.sin_port);
    listen(s, nprocs);


    if (!show_on) {
	signal(SIGHUP, cleanup_handler);
	signal(SIGINT, cleanup_handler);
	signal(SIGTSTP, nostop_handler);
	signal(SIGCHLD, child_handler);
	signal(SIGALRM, alarm_handler);
    }

    alarm(1000);
    /* long timeout for testing, where process may be stopped in debugger */

#ifdef USE_DDD
#define DEBUGGER "/usr/bin/ddd"
#else
#define DEBUGGER "gdb"
#endif

    if (debug_on) {
	char keyval_list[COMMAND_LEN];
	sprintf(keyval_list, "%s", " ");
	/* Take more env variables if present */
	while (strchr(argv[aout_index], '=')) {
	    strcat(keyval_list, argv[aout_index]);
	    strcat(keyval_list, " ");
	    aout_index ++;
	}
	if(use_totalview) {
	    sprintf(command_name_tv, "%s %s %s", keyval_list, 
		    totalview_cmd, argv[aout_index]);
	    sprintf(command_name, "%s %s ", keyval_list, argv[aout_index]);
       	} else {
	    sprintf(command_name, "%s %s %s", keyval_list, 
		    DEBUGGER, argv[aout_index]);
	}
    } else {
        sprintf(command_name, "%s", argv[aout_index]);
    }

    if(use_totalview) {
	/* Only needed for root */
	strcat(command_name_tv, " -a ");
    }

    /* add the arguments */
    for (i = aout_index + 1; i < argc; i++) {
        strcat(command_name, " ");
        strcat(command_name, argv[i]);
    }

   if(use_totalview) {
       /* Complete the command for non-root processes */
   	strcat(command_name, " -mpichtv");

       /* Complete the command for root process */
   	for (i = aout_index + 1; i < argc; i++) {
	    strcat(command_name_tv, " ");
	    strcat(command_name_tv, argv[i]);
 	}
   	strcat(command_name_tv, " -mpichtv");
   }

    /* start all processes */
    for (i = 0; i < nprocs; i++) {
	if((use_totalview) && (i == 0)) {
	    if (start_process(i, command_name_tv, env) < 0) {
	       	fprintf(stderr, 
			"Unable to start process %d on %s. Aborting.\n", 
			i, plist[i].hostname);
	       	cleanup();
	    }
       	} else {
	    if (start_process(i, command_name, env) < 0) {
	       	fprintf(stderr, 
			"Unable to start process %d on %s. Aborting.\n", 
			i, plist[i].hostname);
	       	cleanup(); 
	    } 
	}
    }

    if (show_on)
        exit(0);

    /* accept incoming connections, read port numbers */
    for (i = 0; i < nprocs; i++) {
        int version, rank, nread;
	char pidstr[12];
      ACCEPT:
        sockaddr_len = sizeof(sockaddr);
        s1 = accept(s, (struct sockaddr *) &sockaddr, &sockaddr_len);
        if (s1 < 0) {
            if (errno == EINTR)
                goto ACCEPT;
            perror("accept");
            cleanup();
        }

        /* 
         * protocol: 
         *  0. read protocol version number
         *  1. read rank of process
         *  2. read address length
         *  3. read address itself
         *  4. send array of all addresses
         */


       /* 0. Find out what version of the startup protocol the executable
        * was compiled to use. */

        nread = read(s1, &version, sizeof(version));
        if (nread != sizeof(version)) {
            perror("read");
            cleanup();
        }
        if (version != PMGR_VERSION) {
            fprintf(stderr, "mpirun: executable version %d does not match" 
                  " our version %d.\n", version, PMGR_VERSION);
            cleanup();
        }

        /* 1. Find out who we're talking to */
        nread = read(s1, &rank, sizeof(rank));
        if (nread != sizeof(rank)) {
            perror("read");
            cleanup();
        }

        if (rank < 0 || rank >= nprocs || plist[rank].state != P_STARTED) {
            fprintf(stderr, "mpirun: invalid rank received. \n");
            cleanup();
        }
        plist[rank].control_socket = s1;
        plist[rank].state = P_CONNECTED;
        /* Let us know connection was established
         * printf("MPIRUN_RSH: Process rank %d connected\n",rank);
         */

        /* 2. Find out length of the data */
        nread = read(s1, &addrlen, sizeof(addrlen));
        if (nread != sizeof(addrlen)) {

            /* nread == 0 is not actually an error! */
            if (nread == 0)
                continue;

            perror("read");
            cleanup();
        }

        if (i == 0)
            global_addrlen = addrlen;
        else if (addrlen != global_addrlen) {
            fprintf(stderr, "Address lengths %d and %d do not match\n", 
		    addrlen, global_addrlen);
            cleanup();
        }

	if (addrlen == 0)
	    goto read_pid; 

        if (i == 0) {
            /* allocate as soon as we know the address length */
            alladdrs = (int *) malloc(addrlen * nprocs);
            if (alladdrs == NULL) {
                perror("malloc");
                exit(1);
            }
        }

        /* 3. Read info from each process */

        /* for byte location */
        alladdrs_char = (char *) &alladdrs[rank * addrlen / sizeof(int)];

        tot_nread = 0;

        while (tot_nread < addrlen) {
            nread =
                read(s1, (void *) (alladdrs_char + tot_nread),
                     addrlen - tot_nread);

            if (nread < 0) {
                perror("read");
                cleanup();
            }

            tot_nread += nread;
        }

read_pid:
   	/* 2. Find out length of the data */
	nread = read(s1, &pidlen, sizeof(pidlen));
	if (nread != sizeof(pidlen)) {
	    perror("read");
	    cleanup();
	}

	/*fprintf(stderr, "read Pid lengths %d and %d \n", pidlen, nread);*/
	if (i == 0) {
	    pidglen = pidlen;
	} else {
	    if (pidlen != pidglen) {
		fprintf(stderr, "Pid lengths %d and %d do not match\n", 
			pidlen, pidglen);
		cleanup();
	    }
	}
	    
	if (i == 0) {
	    /* allocate as soon as we know the address length */
	    allpids = (char *)malloc(pidlen*nprocs);
	    if (allpids == NULL) {
		perror("malloc");
		exit(1);
	    }
	}

	tot_nread=0;
	while(tot_nread < pidlen) {
	    nread = read(s1, (void*)(allpids+rank*pidlen+tot_nread), 
		    pidlen - tot_nread);
	    /*fprintf(stderr, "read length %d \n", nread);*/
	    if(nread < 0) {
		perror("read");
		cleanup();
	    }
	    tot_nread += nread;
	}
    }


    /* at this point, all processes have checked in. */

    /* cancel the timeout */
    alarm(0);

    /* send ports to all but highest ranking process, as it needs none */

#ifdef DEBUG
    for (i = 0; i < nprocs; i++) {
        for (j = 0; j < addrlen / sizeof(int); j++) {
            fprintf(stderr, "%d\t",
                    alladdrs[i * addrlen / sizeof(int) + j]);
        }
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
#endif

    out_addrs_len = 3 * nprocs * sizeof(int);
    out_addrs = (int *) malloc(out_addrs_len);
    if (out_addrs == NULL) {
        perror("malloc");
        exit(1);
    }

    for (i = 0; i < nprocs; i++) {
        /* put hca_lid information at the first beginning */
        out_addrs[i] = alladdrs[i * addrlen / sizeof(int) + i];

        /* put host id information in the third round */
        out_addrs[2 * nprocs + i] =
            alladdrs[i * addrlen / sizeof(int) + nprocs];
    }

    for (i = 0; i < nprocs; i++) {
        int nwritten;

        /* personalized address information for each process */
        for (j = 0; j < nprocs; j++) {
            /* put qp information here */
            if (i == j)
                out_addrs[nprocs + j] = -1;
            else
                out_addrs[nprocs + j] =
                    alladdrs[j * addrlen / sizeof(int) + i];
        }

#ifdef DEBUG
        for (j = 0; j < out_addrs_len / sizeof(int); j++)
            fprintf(stderr, "%d\t", out_addrs[j]);
        fprintf(stderr, "\n");
#endif

        nwritten =
            write(plist[i].control_socket, out_addrs, out_addrs_len);
        if (nwritten != out_addrs_len) {
            perror("write");
            cleanup();
        }

	if(pidlen != 0) {
	    nwritten = 0;
	    nwritten = write(plist[i].control_socket, allpids, nprocs*pidlen);
	    if (nwritten != nprocs*pidlen) {
		perror("write");
		cleanup();
	    }
	}

        plist[i].state = P_RUNNING;
    }

#ifndef USE_VIADEV_BARRIER
    /* Not using the VI barrier.  Implement the termination
     * barrier using the socket network we have already
     * established.
     */
    process_termination();

#endif

    /* shut down all our ports */
    close(s);
    for (i = 0; i < nprocs; i++)
        close(plist[i].control_socket);

    /* clients have all information now. Just sit and wait for them
       to die, which we will detect via a signal. 
     */

    while (1) {
        sleep(100);
    }
    exit(0);

}

int start_process(int i, char *command_name, char *env)
{
    char *remote_command;
    char *xterm_command;
    char xterm_title[100];

    int id = getpid();
    int str_len;

    str_len = strlen(command_name) + strlen(env) + strlen(wd) + 
        strlen(mpirun_processes) + 512;

    if ((remote_command = malloc(str_len)) == NULL) {
        fprintf(stderr, "Failed to malloc %d bytes for remote_command\n",
                str_len);
        exit(1);
    }
    if ((xterm_command = malloc(str_len)) == NULL) {
        fprintf(stderr, "Failed to malloc %d bytes for xterm_command\n",
                str_len);
        exit(1);
    }


    /* 
     * this is the remote command we execute whether we were are using 
     * an xterm or using rsh directly 
     */
    sprintf(remote_command, 
	    "cd %s; %s MPIRUN_MPD=0 MPIRUN_HOST=%s MPIRUN_PORT=%d "
	    "MPIRUN_PROCESSES='%s' " 
	    "MPIRUN_RANK=%d MPIRUN_NPROCS=%d MPIRUN_ID=%d %s %s %s",
	    wd, ENV_CMD, mpirun_host, port, mpirun_processes, i, 
	    nprocs, id, display,env, command_name); 

    if (xterm_on) {
        sprintf(xterm_command, "%s; echo process exited", remote_command);
        sprintf(xterm_title, "\"mpirun process %d of %d\"", i, nprocs);
    }

    plist[i].pid = fork();
    plist[i].state = P_STARTED; /* putting after fork() avoids a race */
    if (plist[i].pid == 0) {
        if (i != 0) {
            int fd = open("/dev/null", O_RDWR, 0);
            (void) dup2(fd, STDIN_FILENO);
        }
        if (use_rsh)
            strcpy(sh_cmd, RSH_CMD);
        else
            strcpy(sh_cmd, SSH_CMD);

        if (xterm_on) {
            if (show_on) {
                printf("command: %s -T %s -e %s %s %s %s\n", XTERM,
                       xterm_title, sh_cmd, use_rsh ? "" : SSH_ARG,
                       plist[i].hostname, xterm_command);
            } else {
                if (use_rsh) {
                    execl(XTERM, XTERM, "-T", xterm_title, "-e",
                          sh_cmd, plist[i].hostname, xterm_command, NULL);
                } else {
                    execl(XTERM, XTERM, "-T", xterm_title, "-e",
                          sh_cmd, SSH_ARG, plist[i].hostname,
                          xterm_command, NULL);
                }
            }
        } else {
            if (show_on) {
                printf("command: %s %s %s\n", sh_cmd, plist[i].hostname,
                       remote_command);
            } else {
                if (use_rsh) {
                    execl(sh_cmd, sh_cmd, plist[i].hostname,
                          remote_command, NULL);
                } else {
                    execl(sh_cmd, sh_cmd, SSH_ARG, plist[i].hostname,
                          remote_command, NULL);
                }
            }
        }
        if (!show_on) {
            perror("RSH/SSH command failed!");
        }
        exit(1);
    }

    free(remote_command);
    free(xterm_command);
    return (0);
}

void process_termination()
{
    int i;
    int rval;
    int send_val = 1000;
    int remote_id;

#ifdef MCST_SUPPORT
    int my_rank;
    int mcs_read, mcs_written;
    char grp_info[GRP_INFO_LEN];
    int s;
#endif

    /* we want to wait until all processes have reached a point
     * in their termination where they have cancelled the VIA async
     * error handlers.
     * We implement a simple barrier by waiting in a read on each
     * socket until the process has cancelled its error handler.
     * NOTE: If EOF is returned, this implies the remote process has
     * failed and the connection is broken.
     * Once all processes respond, we inform them to continue by
     * writing a value back down the socket.
     */

    for (i = 0; i < nprocs; i++) {
        int s = plist[i].control_socket;
        int nread;
        remote_id = -1;
        nread = read(s, &remote_id, sizeof(remote_id));
        if (nread == -1) {
            perror("termination socket read failed");
            plist[i].state = P_DISCONNECTED;
        } else if (nread == 0) {
            plist[i].state = P_DISCONNECTED;
        } else if (nread != sizeof(remote_id)) {
            printf("Invalid termination socket read on [%d] "
                   "returned [%d] got [%d]\n", i, nread, remote_id);
            cleanup();
        } else {
            plist[i].state = P_FINISHED;
        }
    }

    /* now, everyone who is still alive has responded */
    for (i = 0; i < nprocs; i++) {
        int s = plist[i].control_socket;
        if (plist[i].state == P_FINISHED) {
            int nwritten = write(s, &send_val, sizeof(send_val));
            if (nwritten != sizeof(send_val)) {
                perror("socket write");
                cleanup();
            }
        }
    }
#ifdef MCST_SUPPORT

    /* First, receive root's group information. Root=0 */

    s = plist[0].control_socket;
    mcs_read = read(s, grp_info, GRP_INFO_LEN);
    if (mcs_read != GRP_INFO_LEN) {
        fprintf(stderr, "mpirun_rsh: read grp info failed "
                "(len %d %s)\n", mcs_read, grp_info);
        cleanup();
    }

    /* Second, send grp info to all others */
    for (i = 1; i < nprocs; i++) {
        s = plist[i].control_socket;
        mcs_written = write(s, grp_info, GRP_INFO_LEN);
        if (mcs_written != GRP_INFO_LEN) {
            fprintf(stderr, "mpirun_rsh: write grp info failed %d\n",
                    mcs_written);
            cleanup();
        }
    }

    /* Third, receive ack from all non-root nodes */
    for (i = 1; i < nprocs; i++) {
        s = plist[i].control_socket;
        mcs_read = read(s, &my_rank, sizeof(my_rank));
        if (mcs_read != sizeof(my_rank)) {
            fprintf(stderr, "mpirun_rsh: write grp info failed %d\n",
                    mcs_written);
            perror("mpirun_rsh: read  grp ack");
            cleanup();
        }
    }

    /* Fourth, write ack to the root */
    my_rank = 0;
    s = plist[0].control_socket;
    mcs_written = write(s, &my_rank, sizeof(my_rank));
    if (mcs_written != sizeof(my_rank)) {
        perror("pmgr_exchange_mcs_group_sync: mpirun_rsh "
               "write ack to root");
        cleanup();
    }
#endif


}

void usage(void)
{
    fprintf(stderr, "usage: mpirun_rsh [-v] [-rsh|-ssh] "
            "[-paramfile=pfile] "
  	    "[-debug] -[tv] [-xterm] [-show] -np N "
            "(-hostfile hfile | h1 h2 ... hN) a.out args\n");
    fprintf(stderr, "Where:\n");
    fprintf(stderr, "\tv          => Show version and exit\n");
    fprintf(stderr, "\trsh        => " "to use rsh for connecting\n");
    fprintf(stderr, "\tssh        => " "to use ssh for connecting\n");
    fprintf(stderr, "\tparamfile  => "
            "file containing run-time MVICH parameters\n");
    fprintf(stderr, "\tdebug      => "
            "run each process under the control of gdb\n");
    fprintf(stderr,"\ttv         => "
	    "run each process under the control of totalview\n");
    fprintf(stderr, "\txterm      => "
            "run remote processes under xterm\n");
    fprintf(stderr, "\tshow       => "
            "show command for remote execution but dont run it\n");
    fprintf(stderr, "\tnp         => "
            "specify the number of processes\n");
    fprintf(stderr, "\th1 h2...   => "
            "names of hosts where processes should run\n");
    fprintf(stderr, "or\thostfile   => "
            "name of file contining hosts, one per line\n");
    fprintf(stderr, "\ta.out      => " "name of MPI binary\n");
    fprintf(stderr, "\targs       => " "arguments for MPI binary\n");
    fprintf(stderr, "\n");

    exit(1);
}

/* finds first non-whitespace char in input string */
char *skip_white(char *s)
{
    /* return pointer to first non-whitespace char in string */
    /* Assumes string is null terminated */
    while ((*s == ' ') || (*s == '\t'))
        s++;
    return s;
}

/*
 * reads the param file and constructs the environment strings
 * for each of the environment variables.
 * The caller is responsible for de-allocating the returned string.
 *
 * NOTE: we cant just append these to our current environment because
 * RSH and SSH do not export our environment vars to the remote host.
 * Rather, the RSH command that starts the remote process looks
 * something like:
 *    rsh remote_host "cd workdir; env ENVNAME=value ... command"
 */
#define ENV_LEN 1024
#define LINE_LEN 256
char *read_param_file(char *paramfile)
{
    FILE *pf;
    char errstr[256];
    char *env;
    char name[128], value[128];
    char buf[384];
    char line[LINE_LEN];
    char *p;
    int num, e_len;
    int env_left = 0;

    if ((pf = fopen(paramfile, "r")) == NULL) {
        sprintf(errstr, "Cant open paramfile = %s", paramfile);
        perror(errstr);
        exit(1);
    }

    if ((env = malloc(ENV_LEN)) == NULL) {
        fprintf(stderr, "Malloc of env failed in read_param_file\n");
        exit(1);
    }
    env_left = ENV_LEN - 1;

    while (fgets(line, LINE_LEN, pf) != NULL) {
        p = skip_white(line);
        if (*p == '#' || *p == '\n') {
            /* a comment or a blank line, ignore it */
            continue;
        }
        /* look for NAME = VALUE, where NAME == MVICH_... */
        name[0] = value[0] = '\0';
        if (param_debug) {
            printf("Scanning: %s\n", p);
        }
        if ((num = sscanf(p, "%64[A-Z_] = %192s", name, value)) != 2) {
            /* debug */
            if (param_debug) {
                printf("FAILED: matched = %d, name = %s, "
                       "value = %s in \n\t%s\n", num, name, value, p);
            }
            continue;
        }

        /* construct the environment string */
        buf[0] = '\0';
        sprintf(buf, "%s=%s ", name, value);
        /* concat to actual environment string */
        e_len = strlen(buf);
        if (e_len > env_left) {
            /* oops, need to grow env string */
            int newlen =
                (ENV_LEN > e_len + 1 ? ENV_LEN : e_len + 1) + strlen(env);
            if ((env = realloc(env, newlen)) == NULL) {
                fprintf(stderr, "realloc failed in read_param_file\n");
                exit(1);
            }
            if (param_debug) {
                printf("realloc to %d\n", newlen);
            }
            env_left = ENV_LEN - 1;
        }
        strcat(env, buf);
        env_left -= e_len;
        if (param_debug) {
            printf("Added: [%s]\n", buf);
            printf("env len = %d, env left = %d\n", strlen(env), env_left);
        }
    }
    fclose(pf);

    return env;
}

void cleanup_handler(int sig)
{
    static int already_called = 0;
    printf("Signal %d received.\n", sig);
    if (already_called == 0) {
        already_called = 1;
    }
    cleanup();

    exit(1);
}

void cleanup(void)
{
    int i;
    /* could walk through list of processes, but it looks
       like we can just send the signal to the process group
     */

    if (use_totalview) {
	fprintf(stderr, "Cleaning up all processes ...");
    }
#ifdef MAC_OSX
    for (i = 0; i < NSIG; i++) {
#else
    for (i = 0; i < _NSIG; i++) {
#endif
        signal(i, SIG_DFL);
    }

    for (i = 0; i < nprocs; i++) {
        if (RUNNING(i)) {
            /* send terminal interrupt, which will hopefully 
               propagate to the other side. (not sure what xterm will
               do here.
             */
            kill(plist[i].pid, SIGINT);
        }
    }
    sleep(1);

    for (i = 0; i < nprocs; i++) {
        if (plist[i].state != P_NOTSTARTED) {
            /* send regular interrupt to rsh */
            kill(plist[i].pid, SIGTERM);
        }
    }

    sleep(1);

    for (i = 0; i < nprocs; i++) {
        if (plist[i].state != P_NOTSTARTED) {
            /* Kill the processes */
            kill(plist[i].pid, SIGKILL);
        }
    }

    fprintf(stderr, "done.\n");

    exit(1);

}


void nostop_handler(int signal)
{
    printf("Stopping from the terminal not allowed\n");
}

void alarm_handler(int signal)
{
    if (use_totalview) {
	fprintf(stderr, "Timeout alarm signaled\n");
    }
    cleanup();
}


void child_handler(int signal)
{
    int status, i, child, pid;
    int exitstatus = 0;

    if (use_totalview) {
	fprintf(stderr, "mpirun: child died. Waiting for others.\n"); 
    }
    alarm(10);
    for (i = 0; i < nprocs; i++) {
        pid = wait(&status);
        if (pid == -1) {
            perror("wait");
            exitstatus = 1;
        } else if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
            exitstatus = 1;
        }
        for (child = 0; child < nprocs; child++) {
            if (plist[child].pid == pid) {
                plist[child].state = P_EXITED;
                break;
            }
        }
        if (child == nprocs) {
            fprintf(stderr, "Unable to find child %d!\n", pid);
            exitstatus = 1;
        }
    }
    alarm(0);
    exit(exitstatus);
}
