
/*
 * Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package, developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL) and headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef VIADEV_RPUT_SUPPORT
#define VIADEV_RPUT_SUPPORT
#endif


#ifdef VIADEV_RPUT_SUPPORT

#include "mpiimpl.h"
#include "mpimem.h"
#include "mpid.h"
#include "mpiops.h"
#include "mpipt2pt.h"

/* For device level initialization */
#ifdef USE_MRAIL
#include "../../mpid/vapi_multirail/collutils.h"
#else
#include "../../mpid/vapi/collutils.h"
#endif

/* Prototype for communicator RDMA features initialization*/
void comm_rdma_init(struct MPIR_COMMUNICATOR*);
void comm_exch_addr(struct MPIR_COMMUNICATOR*);
void set_local_offsets(struct MPIR_COMMUNICATOR* comm);
void coll_prepare_flags(struct Collbuf*);

void comm_rdma_init(struct MPIR_COMMUNICATOR* comm)
{
    struct L_Collbuf* local_coll;
    struct R_Collbuf* remote_coll;

    dreg_entry* buf_entry = NULL;

    if(comm->np <= 1) {
        /* We do not need to do any initialization */

        comm->collbuf = NULL;
        comm->is_alltoall_enabled = 0;
        comm->is_mcast_enabled = 0;
        comm->is_barrier_enabled = 0;
        return;
    }

    comm->collbuf = NULL;

    if(NULL != getenv("DISABLE_RDMA_ALLTOALL")) {
        comm->is_alltoall_enabled = 0;
    } else {

        /* User has enabled Alltoall in general */
        comm->is_alltoall_enabled = 1;
    }

    if(NULL != getenv("DISABLE_HARDWARE_MCST")) {
        /* The user has disabled Multicast */
        comm->is_mcast_enabled = 0;
    } else {
        /* The user has enabled Multicast */
        comm->is_mcast_enabled = 1;
    }

    if(NULL != getenv("DISABLE_RDMA_BARRIER")) {
        /* The user has disabled RDMA Barrier */
        comm->is_barrier_enabled = 0;
    } else {
        /* The user has enabled RDMA Barrier */
        comm->is_barrier_enabled = 1;
    }


    comm->collbuf = (struct Collbuf*)
        malloc(sizeof(struct Collbuf));

    CHECK_ALLOC(comm->collbuf);

    local_coll = (struct L_Collbuf*)
        malloc(sizeof(struct L_Collbuf));

    CHECK_ALLOC(local_coll);

    local_coll->buf = (void*) malloc(COLL_BUF_SIZE);

    CHECK_ALLOC(local_coll->buf);

    /* Set the collective buffer to zero */
    memset(local_coll->buf, 0, COLL_BUF_SIZE);

    /* Register the local memory */
    register_mem(local_coll->buf, COLL_BUF_SIZE,
            &(local_coll->buf_hndl));

    comm->collbuf->l_coll = local_coll;

    /* Allocate space for
     * storing remote buffer
     * addresses */
    remote_coll = (struct R_Collbuf*)
        malloc(sizeof(struct R_Collbuf) * comm->np);

    CHECK_ALLOC(remote_coll);

    /* Set all the memory to zero */
    memset(remote_coll, 0, sizeof(struct R_Collbuf) * comm->np);

    comm->collbuf->r_coll = remote_coll;

    /*Set the flags for rdma barrier */
    comm->collbuf->togle = 0;
    comm->collbuf->rdma_barrier_id = 0;
}

/* Exchange addresses with other processes */
void comm_exch_addr(struct MPIR_COMMUNICATOR* comm)
{
    int i = 0;
    struct Coll_Addr_Exch send_pkt;
    struct Coll_Addr_Exch *recv_pkt;

    MPI_Status* statarray = NULL;
    MPI_Request* reqarray = NULL;

    if(comm->np <= 1) {
        return;
    }


    recv_pkt = (struct Coll_Addr_Exch*)
        malloc(sizeof(struct Coll_Addr_Exch) * comm->np);
    CHECK_ALLOC(recv_pkt);

    statarray = (MPI_Status*) malloc(2 * sizeof(MPI_Status) * comm->np);
    CHECK_ALLOC(statarray);

    reqarray = (MPI_Request*) malloc(2 * sizeof(MPI_Request) * comm->np);
    CHECK_ALLOC(reqarray);

    send_pkt.buf = comm->collbuf->l_coll->buf;
    send_pkt.buf_hndl = comm->collbuf->l_coll->buf_hndl;
    for(i = 0; i < comm->np; i++) {
        /* Don't send to myself */
        if(i == comm->local_rank) continue;

        MPI_Sendrecv((void*)&send_pkt, sizeof(struct Coll_Addr_Exch),
                MPI_BYTE, comm->lrank_to_grank[i], ADDR_EXCHANGE_TAG,
                (void*)&(recv_pkt[i]),sizeof(struct Coll_Addr_Exch),
                MPI_BYTE, comm->lrank_to_grank[i], ADDR_EXCHANGE_TAG,
                MPI_COMM_WORLD, &(statarray[i]));
    }

    for(i = 0; i < comm->np; i++) {
        comm->collbuf->r_coll[i].buf = recv_pkt[i].buf;
        comm->collbuf->r_coll[i].buf_hndl.hndl = recv_pkt[i].buf_hndl.hndl;
        comm->collbuf->r_coll[i].buf_hndl.lkey = recv_pkt[i].buf_hndl.lkey;
        comm->collbuf->r_coll[i].buf_hndl.rkey = recv_pkt[i].buf_hndl.rkey;
    }

    free(reqarray);
    free(statarray);
    free(recv_pkt);
}

void set_local_offsets(struct MPIR_COMMUNICATOR* comm)
{
    struct L_Collbuf* l_coll;
    struct R_Collbuf* r_coll;
    struct Alltoall* atoa = NULL;
    struct Direct_Eager* de = NULL;
    struct DE_Remote_Node* rnode = NULL;
    int i;

    if(comm->np <= 1) {
        return;
    }

    l_coll = comm->collbuf->l_coll;
    r_coll = comm->collbuf->r_coll;

    atoa = (struct Alltoall*) malloc(sizeof(struct Alltoall));
    CHECK_ALLOC(atoa);


    /* Direct Eager stuff */
    de = (struct Direct_Eager*) malloc(sizeof(struct Direct_Eager));
    CHECK_ALLOC(de);

    rnode = (struct DE_Remote_Node*) malloc(sizeof(struct DE_Remote_Node) *
            comm->np);
    CHECK_ALLOC(rnode);

    for(i = 0; i < comm->np; i++) {
        /* Don't set for myself */
        if ( comm->local_rank == i ) continue;

        rnode[i].recv_addr[0] = (void*)
            ((char*)l_coll->buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (i* (sizeof(aint_t) + sizeof(VAPI_rkey_t) + sizeof(char))));

        rnode[i].recv_addr[1] = (void*)
            ((char*)rnode[i].recv_addr[0] +
             (COLL_ADDR_RECV_SPACE/2));

        rnode[i].done[0] = (void*)
            ((char*)l_coll->buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE + COLL_ADDR_RECV_SPACE +
             i*sizeof(char));

        rnode[i].done[1] = (void*)
            ((char*)rnode[i].done[0] + (COLL_ACK_SPACE/2));

        rnode[i].dest_addr[0] = (void*)
            ((char*)r_coll[i].buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE +
             (comm->local_rank*(sizeof(aint_t) + sizeof(VAPI_rkey_t) 
                                + sizeof(char))));

        rnode[i].dest_addr[1] = (void*)
            ((char*)rnode[i].dest_addr[0] +
             (COLL_ADDR_RECV_SPACE/2));

        rnode[i].r_done[0] = (void*)
            ((char*)r_coll[i].buf + COLL_FLAG_SPACE +
             COLL_ADDR_SEND_SPACE + COLL_ADDR_RECV_SPACE +
             (comm->local_rank*sizeof(char)));

        rnode[i].r_done[1] = (void*)
            ((char*)rnode[i].r_done[0] + (COLL_ACK_SPACE/2));

    }

    /* Start with the first set of buffers */
    de->alternate = 0;

    de->node = rnode;

    atoa->de = de;

    comm->collbuf->atoa = atoa;

    /*temp*/
    /* Set the Collective RDMA flag */
    coll_prepare_flags(comm->collbuf);
}

/* A wrap around function to prepare the
 * flags in the beginning of the collective
 * buffer
 */
void coll_prepare_flags(struct Collbuf* coll)
{
    /* Set the first byte to RDMA_FLAG */

    *((char*)coll->l_coll->buf) = ALLTOALL_RDMA_FLAG;

    coll->atoa->rdma_flag = coll->l_coll->buf;

    /* Only a 1 byte long field */
    coll->atoa->rdma_flag_size = 1;
}

#endif /* VIADEV_RPUT_SUPPORT  */
